import os
import h5py
from .hdf5db import Hdf5db
from .jsontoh5 import Writeh5

def dict_to_h5(
        obj,
        h5_path,
        *,
        data_dir=None
    ):
        if "root" not in obj:
            raise Exception("no root key in input file")
        root_uuid = obj["root"]
        
        if os.path.exists(h5_path):
            os.remove(h5_path)

        # create the file, will raise IOError if there's a problem
        Hdf5db.createHDF5File(h5_path) 

        with Hdf5db(h5_path, root_uuid=root_uuid, update_timestamps=False) as db:
            h5writer = Writeh5(db, obj, data_dir=data_dir)
            h5writer.writeFile()

        # open with h5py and remove the _db_ group
        # Note: this will delete any anonymous (un-linked) objects
        f = h5py.File(h5_path, 'a')
        if "__db__" in f:
            del f["__db__"]
        f.close()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
import sys
import json
import argparse
import os.path as op
import tempfile
import numpy as np
import hashlib
import os
from .hdf5dtype import getTypeItem, getItemSize 
from typing import Optional
import random

import logging
import logging.handlers

from .hdf5db import Hdf5db
from .hdf5dtype import getTypeResponse


"""
DumpJson - return json representation of all objects within the given file
"""

class DumpJson:
    def __init__(self, db, app_logger=None, options=None):
        self.options = options
        self.db = db
        if app_logger:
            self.log = app_logger
        else:
            self.log = logging.getLogger()
        self.json = {}

    def dumpAttribute(self, col_name, uuid, attr_name):
        self.log.info("dumpAttribute: [" + attr_name + "]")
        item = self.db.getAttributeItem(col_name, uuid, attr_name)
        response = { 'name': attr_name }
        typeItem = item['type']
        response['type'] = getTypeResponse(typeItem)
        response['shape'] = item['shape']
        # if not self.options.D:
        #     if 'value' not in item:
        #         self.log.warning("no value key in attribute: " + attr_name)
        #     else:
        #         response['value'] = item['value']   # dump values unless header -D was passed
        if 'value' not in item:
            self.log.warning("no value key in attribute: " + attr_name)
        else:
            response['value'] = item['value']   # dump values unless header -D was passed
        return response


    def dumpAttributes(self, col_name, uuid):
        attr_list = self.db.getAttributeItems(col_name, uuid)
        self.log.info("dumpAttributes: " + uuid)
        items = []
        for attr in attr_list:
            item = self.dumpAttribute(col_name, uuid, attr['name'])
            items.append(item)

        return items

    def dumpLink(self, uuid, name):
        item = self.db.getLinkItemByUuid(uuid, name)
        for key in ('ctime', 'mtime', 'href'):
            if key in item:
                del item[key]
        return item


    def dumpLinks(self, uuid):
        link_list = self.db.getLinkItems(uuid)
        items = []
        for link in link_list:
            item = self.dumpLink(uuid, link['title'])
            items.append(item)
        return items


    def dumpGroup(self, uuid):
        item = self.db.getGroupItemByUuid(uuid)
        if 'alias' in item:
            alias = item['alias']
            if alias:
                self.log.info("dumpGroup alias: [" + alias[0] + "]")
        for key in ('ctime', 'mtime', 'linkCount', 'attributeCount', 'id'):
            if key in item:
                del item[key]
        attributes = self.dumpAttributes('groups', uuid)
        if attributes:
            item['attributes'] = attributes
        links = self.dumpLinks(uuid)
        if links:
            item['links'] = links
        return item


    def dumpGroups(self):
        groups = {}
        item = self.dumpGroup(self.root_uuid)
        groups[self.root_uuid] = item
        uuids = self.db.getCollection("groups")
        for uuid in uuids:
            item = self.dumpGroup(uuid)
            groups[uuid] = item

        self.json['groups'] = groups


    def dumpDataset(self, uuid):
        response = { }
        self.log.info("dumpDataset: " + uuid)
        item = self.db.getDatasetItemByUuid(uuid)
        if 'alias' in item:
            alias = item['alias']
            if alias:
                self.log.info("dumpDataset alias: [" + alias[0] + "]")
            response['alias'] = item['alias']
        else:
            alias = None

        typeItem = item['type']
        response['type'] = getTypeResponse(typeItem)
        shapeItem = item['shape']
        shape_rsp = {}
        num_elements = 1
        shape_rsp['class'] = shapeItem['class']
        if 'dims' in shapeItem:
            shape_rsp['dims'] = shapeItem['dims']
            for dim in shapeItem['dims']:
                num_elements *= dim
        if 'maxdims' in shapeItem:
            maxdims = []
            for dim in shapeItem['maxdims']:
                if dim == 0:
                    maxdims.append('H5S_UNLIMITED')
                else:
                    maxdims.append(dim)
            shape_rsp['maxdims'] = maxdims
        response['shape'] = shape_rsp

        if 'creationProperties' in item:
            response['creationProperties'] = item['creationProperties']

        attributes = self.dumpAttributes('datasets', uuid)
        if attributes:
            response['attributes'] = attributes


        shape_class = shapeItem['class']
        # if (not (self.options.D or self.options.d)) or (shape_class == 'H5S_SCALAR'):
        alias_name = None
        if len(alias) > 0:
            if len(alias[0].split('/')) > 0:
                alias_name = alias[0].split('/')[-1]
        include_value = False
        write_binary = False
        include_dataset_names = self.options.get('include_dataset_names', None) or []
        if (self.options.get('include_datasets', False)) or (alias_name in include_dataset_names) or (shape_class == 'H5S_SCALAR'):
            include_value = True
        else:
            dset = self.db.getDatasetObjByUuid(uuid)
            type_info = getTypeItem(dset.dtype)
            type_class = type_info['class']
            if type_class == 'H5T_ARRAY':
                write_binary = True
            elif type_class == 'H5T_FLOAT':
                write_binary = True
            elif type_class == 'H5T_INTEGER':
                write_binary = True
            ###################################################################
            elif type_class == 'H5T_STRING':
                include_value = True
            elif type_class == 'H5T_ENUM':
                include_value = True
            elif type_class == 'H5T_REFERENCE':
                include_value = True
            ###################################################################
            elif type_class == 'H5T_COMPOUND':
                raise Exception('Unsupported type class {}: {}'.format(type_class, alias))
            elif type_class == 'H5T_VLEN':
                raise Exception('Unsupported type class {}: {}'.format(type_class, alias))
            elif type_class == 'H5T_OPAQUE':
                raise Exception('Unsupported type class {}: {}'.format(type_class, alias))
            else:
                raise Exception('Unsupported type class {}: {}'.format(type_class, alias))

        if include_value:
            if num_elements > 0:
                value = self.db.getDatasetValuesByUuid(uuid)
                response['value'] = value   # dump values unless header flag was passed
            else:
                response['value'] = []  # empty list
        if write_binary:
            if self.options['data_dir']:
                print('Writing {} with shape {}'.format(alias, shapeItem.get('dims')))
                fname = self.options['data_dir'] + '/' + _random_string(10) + '.dat.tmp'
                with open(fname, 'wb') as f:
                    self.db.getDatasetValuesByUuid(uuid, format='binary', tofile=f)
                sha1 = _compute_file_hash(path=fname, algorithm='sha1')
                fname2 = self.options['data_dir'] + '/sha1_' + sha1 + '.dat'
                os.rename(fname, fname2)
                response['valueHash'] = dict(
                    sha1=sha1
                )
        return response

    def dumpDatasets(self):
        uuids = self.db.getCollection("datasets")
        if uuids:
            datasets = {}
            for uuid in uuids:
                item = self.dumpDataset(uuid)
                datasets[uuid] = item

            self.json['datasets'] = datasets

    def dumpDatatype(self, uuid):
        response = { }
        item = self.db.getCommittedTypeItemByUuid(uuid)
        response['alias'] = item['alias']
        typeItem = item['type']
        response['type'] = getTypeResponse(typeItem)
        attributes = self.dumpAttributes('datatypes', uuid)
        if attributes:
            response['attributes'] = attributes
        return response


    def dumpDatatypes(self):
        uuids = self.db.getCollection("datatypes")
        if uuids:
            datatypes = {}
            for uuid in uuids:
                item = self.dumpDatatype(uuid)
                datatypes[uuid] = item

            self.json['datatypes'] = datatypes

    def dumpFile(self):

        self.root_uuid = self.db.getUUIDByPath('/')

        db_version_info = self.db.getVersionInfo()

        self.json['apiVersion'] = db_version_info['hdf5-json-version']
        self.json['root'] = self.root_uuid

        self.dumpGroups()

        self.dumpDatasets()

        self.dumpDatatypes()
      
        print(json.dumps(self.json, sort_keys=True, indent=4))

    def toDict(self):
        self.root_uuid = self.db.getUUIDByPath('/')

        db_version_info = self.db.getVersionInfo()

        self.json['apiVersion'] = db_version_info['hdf5-json-version']
        self.json['root'] = self.root_uuid

        self.dumpGroups()

        self.dumpDatasets()

        self.dumpDatatypes()

        return self.json

"""
  Generate a temporary filename to avoid problems with trying to create a dbfile
  in a read-only directory.  (See: https://github.com/HDFGroup/h5serv/issues/37)
"""
def getTempFileName():
    f = tempfile.NamedTemporaryFile(delete=False)
    f.close()
    return f.name


# def main():
#     parser = argparse.ArgumentParser(usage='%(prog)s [-h] [-D|-d] <hdf5_file>')
#     parser.add_argument('-D', action='store_true', help='surpress all data output')
#     parser.add_argument('-d', action='store_true', help='surpress data output for' +
#         ' datasets (but not attribute values)')
#     parser.add_argument('filename', nargs='+', help='HDF5 to be converted to json')
#     args = parser.parse_args()

#     # create logger
#     log = logging.getLogger("h5serv")
#     # log.setLevel(logging.WARN)
#     log.setLevel(logging.INFO)
#     # add log handler
#     handler = logging.FileHandler('./h5tojson.log')

#     # add handler to logger
#     log.addHandler(handler)

#     filename = args.filename[0]
#     if not op.isfile(filename):
#         sys.exit("Cannot find file: %s" % filename)

#     log.info("h5tojson " + filename)

#     dbFilename = getTempFileName()
#     log.info("Using dbFile: " + dbFilename)
#     with Hdf5db(filename, dbFilePath=dbFilename, readonly=True, app_logger=log) as db:
#         dumper = DumpJson(db, app_logger=log, options=args)
#         dumper.dumpFile()

def _write_single_dataset_hdf5(path, dset):
    import h5py
    with h5py.File(path, 'w') as f:
        f.create_dataset('dataset', data=dset)

def _compute_file_hash(path: str, algorithm: str) -> Optional[str]:
    if not os.path.exists(path):
        return None
    if (os.path.getsize(path) > 1024 * 1024 * 100):
        print('Computing {} of {}'.format(algorithm, path))
    BLOCKSIZE = 65536
    hashsum = getattr(hashlib, algorithm)()
    with open(path, 'rb') as file:
        buf = file.read(BLOCKSIZE)
        while len(buf) > 0:
            hashsum.update(buf)
            buf = file.read(BLOCKSIZE)
    return hashsum.hexdigest()

def _random_string(num_chars: int) -> str:
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
    return ''.join(random.choice(chars) for _ in range(num_chars))

# if __name__ == "__main__":
#     main()



import os
from .hdf5db import Hdf5db
from .h5tojson import DumpJson

def h5_to_dict(
        h5_path,
        *,
        include_datasets=False,
        include_dataset_names=[],
        data_dir=None
    ):
        """
        Generate a temporary filename to avoid problems with trying to create a dbfile
        in a read-only directory.  (See: https://github.com/HDFGroup/h5serv/issues/37)
        """
        if data_dir is not None:
            if not os.path.exists(data_dir):
                os.mkdir(data_dir)

        import tempfile
        def getTempFileName():
            f = tempfile.NamedTemporaryFile(delete=False)
            f.close()
            return f.name

        if not os.path.exists('_datasets'):
            os.mkdir('_datasets')

        dbFilename = getTempFileName()
        options = dict(
            include_datasets=include_datasets,
            include_dataset_names=include_dataset_names,
            data_dir=data_dir
        )
        with Hdf5db(h5_path, dbFilePath=dbFilename, readonly=True) as db:
            dumper = DumpJson(db, options=options)
            X = dumper.toDict()
        return X


def hierarchy(X):
    nodes_by_alias = dict()
    groups = X.get('groups', {})
    for group in groups.values():
        aliases = group.get('alias', [])
        for alias0 in aliases:
            node = _find_or_create_node(nodes_by_alias, alias0)
            node['_attributes'] = _make_attributes(group.get('attributes', []))
    datasets = X.get('datasets', {})
    for dataset in datasets.values():
        aliases = dataset.get('alias', [])
        for alias0 in aliases:
            ind = alias0.rfind('/')
            if ind >= 0:
                parent_alias = alias0[:ind] or '/'
                name = alias0[ind + 1:]
                parent_node = _find_or_create_node(nodes_by_alias, parent_alias)
                if '_datasets' not in parent_node:
                    parent_node['_datasets'] = dict()
                dsnode = dict(
                    _attributes=_make_attributes(dataset.get('attributes', [])),
                )
                for key in ['type', 'shape', 'creationProperties', 'value', 'valueHash']:
                    if key in dataset:
                        dsnode[key] = dataset[key]
                parent_node['_datasets'][name] = dsnode
    return dict(
        root=nodes_by_alias.get('/', None)
    )

def _find_or_create_node(nodes_by_alias, alias):
    if alias in nodes_by_alias:
        return nodes_by_alias[alias]
    node = dict()
    nodes_by_alias[alias] = node
    if alias != '/':
        ind = alias.rfind('/')
        if ind >= 0:
            parent_alias = alias[:ind] or '/'
            name = alias[ind + 1:]
            parent_node = _find_or_create_node(nodes_by_alias, parent_alias)
            parent_node[name] = node
    return node

def _make_attributes(attributes_list):
    ret = dict()
    for a in attributes_list:
        ret[a['name']] = a
    return ret


##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
This class is used to map between HDF5 type representations and numpy types

"""
import numpy as np
from h5py.h5t import special_dtype
from h5py.h5t import check_dtype
from h5py.h5r import Reference
from h5py.h5r import RegionReference
import six

if six.PY3:
    unicode = str


"""
Convert the given type item  to a predefined type string for
predefined integer and floating point types ("H5T_STD_I64LE", et. al).
For compound types, recursively iterate through the typeItem and do same
conversion for fields of the compound type.
"""
def getTypeResponse(typeItem):

    response = None
    if 'uuid' in typeItem:
        # committed type, just return uuid
        response = 'datatypes/' + typeItem['uuid']
    elif typeItem['class'] == 'H5T_INTEGER' or  typeItem['class'] == 'H5T_FLOAT':
        # just return the class and base for pre-defined types
        response = {}
        response['class'] = typeItem['class']
        response['base'] = typeItem['base']
    elif typeItem['class'] == 'H5T_OPAQUE':
        response = {}
        response['class'] = 'H5T_OPAQUE'
        response['size'] = typeItem['size']
    elif typeItem['class'] == 'H5T_REFERENCE':
        response = {}
        response['class'] = 'H5T_REFERENCE'
        response['base'] = typeItem['base']
    elif typeItem['class'] == 'H5T_COMPOUND':
        response = {}
        response['class'] = 'H5T_COMPOUND'
        fieldList = []
        for field in typeItem['fields']:
            fieldItem = { }
            fieldItem['name'] = field['name']
            fieldItem['type'] = getTypeResponse(field['type'])  # recursive call
            fieldList.append(fieldItem)
        response['fields'] = fieldList
    else:
        response = {}   # otherwise, return full type
        for k in typeItem.keys():
            if k == 'base':
                if type(typeItem[k]) == dict:
                    response[k] = getTypeResponse(typeItem[k])  # recursive call
                else:
                    response[k] = typeItem[k]  # predefined type
            elif k not in ('size', 'base_size'):
                response[k] = typeItem[k]
    return response
    
    
"""
    Get size of an item in bytes.
    For variable length types (e.g. variable length strings),
    return the string "H5T_VARIABLE"
"""
def getItemSize(typeItem):
    # handle the case where we are passed a primitive type first
    if type(typeItem) in [six.string_types, six.text_type, six.binary_type]:
        for type_prefix in ("H5T_STD_I", "H5T_STD_U", "H5T_IEEE_F"): 
            if typeItem.startswith(type_prefix):
                num_bits = typeItem[len(type_prefix):]
                if num_bits[-2:] in ('LE', 'BE'):
                    num_bits = num_bits[:-2]
                try:
                    return int(num_bits) // 8
                except ValueError:
                    raise TypeError("Invalid Type")
        # none of the expect primative types mathched
        raise TypeError("Invalid Type")   
    if type(typeItem) != dict:
        raise TypeError("invalid type")

    item_size = 0
    if 'class' not in typeItem:
        raise KeyError("'class' not provided")
    typeClass = typeItem['class']

       
    if typeClass == 'H5T_INTEGER':
        if 'base' not in typeItem:
            raise KeyError("'base' not provided")
        item_size = getItemSize(typeItem['base'])
         
    elif typeClass == 'H5T_FLOAT':
        if 'base' not in typeItem:
            raise KeyError("'base' not provided")
        item_size = getItemSize(typeItem['base'])
        
    elif typeClass == 'H5T_STRING':
        if 'length' not in typeItem:
            raise KeyError("'length' not provided")
        item_size = typeItem["length"]    
          
    elif typeClass == 'H5T_VLEN':
        item_size = "H5T_VARIABLE"
    elif typeClass == 'H5T_OPAQUE':
        if 'size' not in typeItem:
            raise KeyError("'size' not provided")
        item_size = int(typeItem['size'])
        
    elif typeClass == 'H5T_ARRAY':
        if 'dims' not in typeItem:
            raise KeyError("'dims' must be provided for array types")
        if 'base' not in typeItem:
            raise KeyError("'base' not provided")
        item_size = getItemSize(typeItem['base'])
        
    elif typeClass == 'H5T_ENUM':
        if 'base' not in typeItem:
            raise KeyError("'base' must be provided for enum types")
        item_size = getItemSize(typeItem['base'])
        
    elif typeClass == 'H5T_REFERENCE':
        item_size = "H5T_VARIABLE"
    elif typeClass == 'H5T_COMPOUND':
        if 'fields' not in typeItem:
            raise KeyError("'fields' not provided for compound type")
        fields = typeItem['fields']
        if type(fields) is not list:
            raise TypeError("Type Error: expected list type for 'fields'")
        if not fields:
            raise KeyError("no 'field' elements provided")
        # add up the size of each sub-field
        for field in fields:
            if type(field) != dict:
                raise TypeError("Expected dictionary type for field")
            if 'type' not in field:
                raise KeyError("'type' missing from field")
            subtype_size = getItemSize(field['type']) # recursive call
            if subtype_size == "H5T_VARIABLE":
                item_size = "H5T_VARIABLE"
                break  # don't need to look at the rest
                
            item_size += subtype_size           
    else:
        raise TypeError("Invalid type class")
     
    # calculate array type    
    if 'dims' in typeItem and type(item_size) is int:
        dims = typeItem['dims']
        for dim in dims:
            item_size *= dim   
                
    return item_size



"""
    Return type info.
          For primitive types, return string with typename
          For compound types return array of dictionary items
"""
def getTypeItem(dt):
     
    predefined_int_types = {
        'int8':    'H5T_STD_I8',
        'uint8':   'H5T_STD_U8',
        'int16':   'H5T_STD_I16',
        'uint16':  'H5T_STD_U16',
        'int32':   'H5T_STD_I32',
        'uint32':  'H5T_STD_U32',
        'int64':   'H5T_STD_I64',
        'uint64':  'H5T_STD_U64'
    }
    predefined_float_types = {
        'float32': 'H5T_IEEE_F32',
        'float64': 'H5T_IEEE_F64'
    }
    
    type_info = {}
    if len(dt) > 1:
        # compound type
        names = dt.names
        type_info['class'] = 'H5T_COMPOUND'
        fields = []
        for name in names:
            field = { 'name': name }
            field['type'] = getTypeItem(dt[name])
            fields.append(field)
            type_info['fields'] = fields
    elif dt.shape:
        # array type
        if dt.base == dt:
            raise TypeError("Expected base type to be different than parent")
        # array type
        type_info['dims'] = dt.shape
        type_info['class'] = 'H5T_ARRAY'
        type_info['base'] = getTypeItem(dt.base)
    elif dt.kind == 'O':
        # vlen string or data
        #
        # check for h5py variable length extension
        vlen_check = check_dtype(vlen=dt.base)
        if vlen_check is not None and type(vlen_check) != np.dtype:
            vlen_check = np.dtype(vlen_check)
        ref_check = check_dtype(ref=dt.base)
        if vlen_check == six.binary_type:
            type_info['class'] = 'H5T_STRING'
            type_info['length'] = 'H5T_VARIABLE'
            type_info['charSet'] = 'H5T_CSET_ASCII'
            type_info['strPad'] = 'H5T_STR_NULLTERM'
        elif vlen_check == six.text_type:
            type_info['class'] = 'H5T_STRING'
            type_info['length'] = 'H5T_VARIABLE'
            type_info['charSet'] = 'H5T_CSET_UTF8'
            type_info['strPad'] = 'H5T_STR_NULLTERM'
        elif type(vlen_check) == np.dtype:
            # vlen data
            type_info['class'] = 'H5T_VLEN'
            type_info['size'] = 'H5T_VARIABLE'
            type_info['base'] = getTypeItem(vlen_check)
        elif vlen_check is not None:
            #unknown vlen type
            raise TypeError("Unknown h5py vlen type: " + str(vlen_check))
        elif ref_check is not None:
            # a reference type
            type_info['class'] = 'H5T_REFERENCE'

            if ref_check is Reference:
                type_info['base'] = 'H5T_STD_REF_OBJ'  # objref
            elif ref_check is RegionReference:
                type_info['base'] = 'H5T_STD_REF_DSETREG'  # region ref
            else:
                raise TypeError("unexpected reference type")
        else:
            raise TypeError("unknown object type")
    elif dt.kind == 'V':
        # void type
        type_info['class'] = 'H5T_OPAQUE'
        type_info['size'] = dt.itemsize
        type_info['tag'] = ''  # todo - determine tag
    elif dt.base.kind == 'S':
        # Fixed length string type
        type_info['class'] = 'H5T_STRING'
        type_info['charSet'] = 'H5T_CSET_ASCII'
        type_info['length'] = dt.itemsize
        type_info['strPad'] = 'H5T_STR_NULLPAD'
    elif dt.base.kind == 'U':
        # Fixed length unicode type
        raise TypeError("Fixed length unicode type is not supported")
         
    elif dt.kind == 'b':
        # boolean type - h5py stores as enum
        # assume LE unless the numpy byteorder is '>'
        byteorder = 'LE'
        if dt.base.byteorder == '>':
            byteorder = 'BE'
        # this mapping is an h5py convention for boolean support
        mapping =  {
            "FALSE": 0,
            "TRUE": 1
        }      
        type_info['class'] = 'H5T_ENUM'
        type_info['mapping'] = mapping
        base_info = { "class": "H5T_INTEGER" }
        base_info['base'] = "H5T_STD_I8" + byteorder
        type_info["base"] = base_info

    elif dt.kind == 'f':
        # floating point type
        type_info['class'] = 'H5T_FLOAT'
        byteorder = 'LE'
        if dt.byteorder == '>':
            byteorder = 'BE'
        if dt.name in predefined_float_types:
            #maps to one of the HDF5 predefined types
            type_info['base'] = predefined_float_types[dt.base.name] + byteorder
        else:
            raise TypeError("Unexpected floating point type: " + dt.name)
    elif dt.kind == 'i' or dt.kind == 'u':
        # integer type
        
        # assume LE unless the numpy byteorder is '>'
        byteorder = 'LE'
        if dt.base.byteorder == '>':
            byteorder = 'BE'
             
        # numpy integer type - but check to see if this is the hypy
        # enum extension
        mapping = check_dtype(enum=dt)

        if mapping:
            # yes, this is an enum!
            type_info['class'] = 'H5T_ENUM'
            type_info['mapping'] = mapping
            if dt.name not in predefined_int_types:
                raise TypeError("Unexpected integer type: " + dt.name)
            #maps to one of the HDF5 predefined types
            base_info = { "class": "H5T_INTEGER" }
            base_info['base'] = predefined_int_types[dt.name] + byteorder
            type_info["base"] = base_info
        else:
            type_info['class'] = 'H5T_INTEGER'
            base_name = dt.name
            
            if dt.name not in predefined_int_types:
                raise TypeError("Unexpected integer type: " + dt.name)
           
            type_info['base'] = predefined_int_types[base_name] + byteorder
             
    else:
        # unexpected kind
        raise TypeError("unexpected dtype kind: " + dt.kind)
            
                 
    return type_info


def getNumpyTypename(hdf5TypeName, typeClass=None):
    predefined_int_types = {
          'H5T_STD_I8':  'i1',
          'H5T_STD_U8':  'u1',
          'H5T_STD_I16': 'i2',
          'H5T_STD_U16': 'u2',
          'H5T_STD_I32': 'i4',
          'H5T_STD_U32': 'u4',
          'H5T_STD_I64': 'i8',
          'H5T_STD_U64': 'u8'
    }
    predefined_float_types = {
          'H5T_IEEE_F32': 'f4',
          'H5T_IEEE_F64': 'f8'
    }

    if len(hdf5TypeName) < 3:
        raise Exception("Type Error: invalid typename: ")
    endian = '<'  # default endian
    key = hdf5TypeName
    if hdf5TypeName.endswith('LE'):
        key = hdf5TypeName[:-2]
    elif hdf5TypeName.endswith('BE'):
        key = hdf5TypeName[:-2]
        endian = '>'

    if key in predefined_int_types and (typeClass == None or
            typeClass == 'H5T_INTEGER'):
        return endian + predefined_int_types[key]
    if key in predefined_float_types and (typeClass == None or
            typeClass == 'H5T_FLOAT'):
        return endian + predefined_float_types[key]
    raise TypeError("Type Error: invalid type")


def createBaseDataType(typeItem):

    dtRet = None
    if type(typeItem) in (str, unicode):
        # should be one of the predefined types
        dtName = getNumpyTypename(typeItem)
        dtRet = np.dtype(dtName)
        return dtRet  # return predefined type

    if type(typeItem) != dict:
        raise TypeError("Type Error: invalid type")

    if 'class' not in typeItem:
        raise KeyError("'class' not provided")
    typeClass = typeItem['class']
       
    if typeClass == 'H5T_INTEGER':
        if 'base' not in typeItem:
            raise KeyError("'base' not provided")
        if 'dims' in typeItem:
            raise TypeError("'dims' not supported for integer types")
        baseType = getNumpyTypename(typeItem['base'], typeClass='H5T_INTEGER')
        dtRet = np.dtype(baseType)
    elif typeClass == 'H5T_FLOAT':
        if 'base' not in typeItem:
            raise KeyError("'base' not provided")
        if 'dims' in typeItem:
            raise TypeError("'dims' not supported for floating point types")
        baseType = getNumpyTypename(typeItem['base'], typeClass='H5T_FLOAT')
        dtRet = np.dtype(baseType)
    elif typeClass == 'H5T_STRING':
        if 'length' not in typeItem:
            raise KeyError("'length' not provided")
        if 'charSet' not in typeItem:
            raise KeyError("'charSet' not provided")

        if typeItem['length'] == 'H5T_VARIABLE':
            if 'dims' in typeItem:
                raise TypeError("'dims' not supported for variable types")
            if typeItem['charSet'] == 'H5T_CSET_ASCII':
                dtRet = special_dtype(vlen=bytes)
            elif typeItem['charSet'] == 'H5T_CSET_UTF8':
                dtRet = special_dtype(vlen=unicode)
            else:
                raise TypeError("unexpected 'charSet' value")
        else:
            nStrSize = typeItem['length']
            if type(nStrSize) != int:
                raise TypeError("expecting integer value for 'length'")
            type_code = None
            if typeItem['charSet'] == 'H5T_CSET_ASCII':
                type_code = 'S'
            elif typeItem['charSet'] == 'H5T_CSET_UTF8':
                raise TypeError("fixed-width unicode strings are not supported")
            else:
                raise TypeError("unexpected 'charSet' value")
            dtRet = np.dtype(type_code + str(nStrSize))  # fixed size string
    elif typeClass == 'H5T_VLEN':
        if 'dims' in typeItem:
            raise TypeError("'dims' not supported for vlen types")
        if 'base' not in typeItem:
            raise KeyError("'base' not provided")
        baseType = createBaseDataType(typeItem['base'])
        dtRet = special_dtype(vlen=np.dtype(baseType))
    elif typeClass == 'H5T_OPAQUE':
        if 'dims' in typeItem:
            raise TypeError("'dims' not supported for opaque types")
        if 'size' not in typeItem:
            raise KeyError("'size' not provided")
        nSize = int(typeItem['size'])
        if nSize <= 0:
            raise TypeError("'size' must be non-negative")
        dtRet = np.dtype('V' + str(nSize))
    elif typeClass == 'H5T_ARRAY':
        if not 'dims' in typeItem:
            raise KeyError("'dims' must be provided for array types")
        if 'base' not in typeItem:
            raise KeyError("'base' not provided")
        arrayBaseType = typeItem['base']
        if type(arrayBaseType) is dict:
            if "class" not in arrayBaseType:
                raise KeyError("'class' not provided for array base type")
            if arrayBaseType["class"] not in ('H5T_INTEGER', 'H5T_FLOAT', 'H5T_STRING'):
                raise TypeError("Array Type base type must be integer, float, or string")

        dt_base = createDataType(arrayBaseType)
        
        if type(typeItem['dims']) == int:
            dims = (typeItem['dims'])  # make into a tuple
        elif type(typeItem['dims']) not in (list, tuple):
            raise TypeError("expected list or integer for dims")
        else:
            dims = typeItem['dims']
        # create an array type of the base type
         
        dtRet = np.dtype((dt_base, dims))
        
       
    elif typeClass == 'H5T_REFERENCE':
        if 'dims' in typeItem:
            raise TypeError("'dims' not supported for reference types")
        if 'base' not in typeItem:
            raise KeyError("'base' not provided")
        if typeItem['base'] == 'H5T_STD_REF_OBJ':
        	dtRet = special_dtype(ref=Reference)
        elif typeItem['base'] == 'H5T_STD_REF_DSETREG':
        	dtRet = special_dtype(ref=RegionReference)
        else:
            raise TypeError("Invalid base type for reference type")
    elif typeClass == 'H5T_ENUM':
        if 'base' not in typeItem:
            raise KeyError("Expected 'base' to be provided for enum type")
        base_json = typeItem["base"]
        if 'class' not in base_json:
            raise KeyError("Expected class field in base type")
        if base_json['class'] != 'H5T_INTEGER':
            raise TypeError("Only integer base types can be used with enum type")
        if 'mapping' not in typeItem:
            raise KeyError("'mapping' not provided for enum type")
        mapping = typeItem["mapping"]
        if len(mapping) == 0:
            raise KeyError("empty enum map")
        
        dt = createBaseDataType(base_json)
        if dt.kind == 'i' and dt.name=='int8' and len(mapping) == 2 and 'TRUE' in mapping and 'FALSE' in mapping:
            # convert to numpy boolean type
            dtRet = np.dtype("bool")
        else:
            # not a boolean enum, use h5py special dtype 
            dtRet = special_dtype(enum=(dt, mapping))

    else:
        raise TypeError("Invalid type class")
    

    return dtRet

"""
Create a numpy datatype given a json type 
"""
def createDataType(typeItem):
    dtRet = None
    if type(typeItem) in [six.string_types, six.text_type, six.binary_type]:
        # should be one of the predefined types
        dtName = getNumpyTypename(typeItem)
        dtRet = np.dtype(dtName)
        return dtRet  # return predefined type

    if type(typeItem) != dict:
        raise TypeError("invalid type")


    if 'class' not in typeItem:
        raise KeyError("'class' not provided")
    typeClass = typeItem['class']

    if typeClass == 'H5T_COMPOUND':
        if 'fields' not in typeItem:
            raise KeyError("'fields' not provided for compound type")
        fields = typeItem['fields']
        if type(fields) is not list:
            raise TypeError("Type Error: expected list type for 'fields'")
        if not fields:
            raise KeyError("no 'field' elements provided")
        subtypes = []
        for field in fields:

            if type(field) != dict:
                raise TypeError("Expected dictionary type for field")
            if 'name' not in field:
                raise KeyError("'name' missing from field")
            if 'type' not in field:
                raise KeyError("'type' missing from field")
            field_name = field['name']
            if type(field_name) == unicode:
                # verify the field name is ascii
                try:
                    ascii_name = field_name.encode('ascii')
                except UnicodeDecodeError:
                    raise TypeError("non-ascii field name not allowed")
                if not six.PY3:
                    field['name'] = ascii_name
             
            dt = createDataType(field['type'])  # recursive call
            if dt is None:
                raise Exception("unexpected error")
            subtypes.append((field['name'], dt))  # append tuple
            
        dtRet = np.dtype(subtypes)
        
    else:
        dtRet = createBaseDataType(typeItem)  # create non-compound dt
    return dtRet



import numpy as np
import six
from .hdf5dtype import createDataType

if six.PY3:
    unicode = str

def get_value(X, data_dir=None):
    body = X

    ## note: this is duplicated code. will remedy in future ###########################
    datatype = body['type']
    if type(datatype) in (str, unicode) and datatype.startswith("datatypes/"):
        #committed datatype, just pass in the UUID part
        datatype = datatype[len("datatypes/"):]
    dims = ()  # if no space in body, default to scalar
    max_shape=None
    fill_value=None
    creation_props=None
    if 'creationProperties' in body:
        creation_props = body['creationProperties']
    if "shape" in body:
        shape = body["shape"]
        if shape["class"] == 'H5S_SIMPLE':
            dims = shape["dims"]
            if type(dims) == int:
                # convert int to array
                dim1 = shape
                dims = [dim1]
            if "maxdims" in shape:
                max_shape = shape["maxdims"]
                if type(max_shape) == int:
                    #convert to array
                    dim1 = max_shape
                    max_shape = [dim1]
                # convert H5S_UNLIMITED's to None's
                for i in range(len(max_shape)):
                    if max_shape[i] == 'H5S_UNLIMITED':
                        max_shape[i] = None
        elif shape["class"] == 'H5S_NULL':
            dims = None
    
    dt = createDataType(datatype)

    if "value" in body:
        value0 = body["value"]
        if dims is not None:
            return np.array(value0, dtype=dt).reshape(tuple(dims))
        else:
            return value0
    elif "valueHash" in body:
        if dims is None:
            raise Exception('This case not yet handled.')
        if not data_dir:
            raise Exception('Cannot retrieve data from valueHash because data_dir is not specified')
        fname = None
        for alg in ['sha1', 'md5']:
            if alg in body['valueHash']:
                fname = data_dir + '/{}_{}.dat'.format(alg, body['valueHash'][alg])
        if fname is None:
            raise Exception('Unexpected valueHash')
        return np.memmap(fname, dtype=dt, shape=tuple(dims))
